"""Isotopologue correction using statistical distributions.

Read raw data as csv or excel file and correct for natural abundance.

Usage:
    picor FILE MOLECULE [-o OUTFILE]
                        [-s COL]...
                        [-x EXCOL]...
                        [--isotopes-file IFILE]
                        [--molecules-file MFILE]
                        [-v | -vv]
    picor FILE MOLECULE --res-correction
                        [--mz-calibration MZ]
                        [--isotopes-file IFILE]
                        [-o OUTFILE]
                        [-s COL]...
                        [-x EXCOL]...
                        [--resolution RES]
                        [--molecules-file MFILE]
                        [-v | -vv]
    picor (-h | --help)
    picor --version

Arguments:
    FILE       Path to excel or csv file (xlsx or csv)
    MOLECULE   Name as in molecules-file

Options:
  -o --output OUTFILE       Output file path (csv format)
  -r --res-correction       Perform resolution correction
  --resolution RES          Resolution of measurement [default: 60000]
  --mz-calibration MZ       Mass-charge of calibration point [default: 200]
  -s --subset COL           Column for calculation; can be used multiple times
  -x --exclude-col EXCOL    Column to ignore; can be used multiple times
  --molecules-file MFILE    Path to tab-separated molecules file
                            Name, formula and charge as rows; e.g. Suc C4H4O3 -1
  --isotopes-file IFILE     Path to tab-separated isotope file
                            Element, mass, abundance and isotope as rows
                            E.g. H 1.008 0.99 H01
  -v --verbose              Display more info like intermediate correction factors
                            (multiple increase verbosity, up to 2)
                            Prints to stderr
  -h --help                 Show this screen.
  --version                 Show version.

"""
import logging
from pathlib import Path
import sys
from importlib.metadata import version

from docopt import docopt
import pandas as pd

import picor

__author__ = "Jørn Dietze"
__copyright__ = "Jørn Dietze"
__license__ = "gpl3"

_logger = logging.getLogger(__name__)


def cli(arguments):
    """Run isotope correction with CLI interface."""
    infile = Path(arguments["FILE"])
    outfile = arguments["--output"]
    if infile.suffix == ".xlsx":
        raw_data = pd.read_excel(infile, index_col=0, engine="openpyxl")
    elif infile.suffix == ".csv":
        raw_data = pd.read_csv(infile, index_col=0)
    else:
        raise ValueError("FILE can be either '.csv' or '.xlsx' file type")
    if arguments["--verbose"] == 0:
        logging_level = "WARNING"
    elif arguments["--verbose"] == 1:
        logging_level = "INFO"
    else:
        logging_level = "DEBUG"
    corr_data = picor.calc_isotopologue_correction(
        raw_data,
        arguments["MOLECULE"],
        subset=arguments["--subset"],
        exclude_col=arguments["--exclude-col"],
        resolution_correction=arguments["--res-correction"],
        mz_calibration=float(arguments["--mz-calibration"]),
        resolution=float(arguments["--resolution"]),
        isotopes_file=arguments["--isotopes-file"],
        molecules_file=arguments["--molecules-file"],
        logging_level=logging_level,
    )
    if outfile:
        corr_data.to_csv(outfile)
    print(corr_data)


def main():
    """Serve as entry point for CLI."""
    package_version = version("picor")
    arguments = docopt(__doc__, version=package_version)
    _logger.info("arguments = {arguments}")
    sys.exit(cli(arguments))


if __name__ == "__main__":
    main()



"""Isotopologue correction using statistical distributions.

Functions:
    calc_indirect_overlap_prob: probability of random H02 or C13
    warn_direct_overlap: warn if direct overlap due to resolution
    warn_indirect_overlap: warn if overlap due to H02 or C13 incorporation
"""
import itertools
import logging
import warnings

import pandas as pd

from picor.isotope_probabilities import calc_label_diff_prob, Label

__author__ = "Jørn Dietze"
__copyright__ = "Jørn Dietze"
__license__ = "gpl3"

_logger = logging.getLogger(__name__)


class ResolutionCorrectionInfo:
    """Class to store all parameters for resolution correction."""

    def __init__(self, do_correction, resolution, mz_calibration, molecule_info):
        self.do_correction = bool(do_correction)
        self.resolution = float(resolution)
        self.mz_calibration = float(mz_calibration)
        self.molecule_info = molecule_info
        self.molecule_mass = Label("No label", molecule_info).calc_isotopologue_mass()
        self.min_mass_diff = self.calc_min_mass_diff(
            self.molecule_mass,
            molecule_info.charge,
            mz_calibration,
            resolution,
        )
        # TODO Replace hardcoded  indirect_overlap_cutoff with flexible approach
        self.indirect_overlap_cutoff = 5

    def __repr__(self):
        return (
            "ResolutionCorrectionInfo("
            f"do_correction={self.do_correction}, resolution={self.resolution}, "
            f"mz_calibration={self.mz_calibration}, molecule_info={self.molecule_info})"
        )

    @staticmethod
    def fwhm(mz_cal, mz, resolution):
        """Calculate Full width half maximum (FWHM)."""
        if mz_cal < 0 or mz <= 0 or resolution <= 0:
            raise ValueError("Arguments must be positive")
        return mz ** (3 / 2) / (resolution * mz_cal**0.5)

    @staticmethod
    def calc_min_mass_diff(mass, charge, mz_cal, resolution):
        """Calculate minimal resolvable mass difference.

        For m and z return minimal mass difference that ca be resolved properly.

        Parameters
        ----------
        mass : float
            Mass of molecule
        charge : int
            Charge of molecule
        mz_cal : float
            Mass/charge for which resolution was determined
        resolution : float
            Resolution

        Returns
        -------
        float
            Minimal resolvable mass differenceo

        Raises
        ------
        ValueError
            If molecule mass is negative.
        """
        if mass < 0:
            raise ValueError("'mass' must be positive.")
        if charge == 0:
            raise ValueError("'charge' cannot be 0, change in metabolites.csv file is required.")
        mz = abs(mass / charge)
        return (
            1.66 * abs(charge) * ResolutionCorrectionInfo.fwhm(mz_cal, mz, resolution)
        )


def calc_coarse_mass_difference(label1, label2):
    """Calculate difference in nucleons (e.g. 2 between H20 and D20).

    nucleons(label2) - nucleons(label1)

    Parameters
    ----------
    label1 : Label
        Label of isotopologue 1
    label2 : Label
        Label of isotopologue 2

    Returns
    -------
    float
        Coarse difference

    Raises
    ------
    TypeError
        If label1 or label2 is not Label instance
    """
    if not isinstance(label1, Label) or not isinstance(label2, Label):
        raise TypeError("label1 and label2 have to be Label instances")
    return label2.mass - label1.mass


def is_isotologue_overlap(
    label1,
    label2,
    res_corr_info,
):
    """Return True if label1 and label2 are too close to detection limit.

    Checks whether two isotopologues defined by label and metabolite name
    are below miniumum resolved mass difference.
    """
    mass1 = label1.calc_isotopologue_mass()
    mass2 = label2.calc_isotopologue_mass()
    return abs(mass1 - mass2) < res_corr_info.min_mass_diff


def warn_indirect_overlap(label_list, res_corr_info):
    """Warn if any of labels can have indirect overlap."""
    for label1, label2 in itertools.permutations(label_list, 2):
        prob = calc_indirect_overlap_prob(label1, label2, res_corr_info)
        if prob:
            warning = f"{label1} indirect overlap with {label2} with prob {prob:.4f}"
            warnings.warn(warning)
            _logger.warning(warning)


def calc_indirect_overlap_prob(label1, label2, res_corr_info):
    """Calculate probability for overlap caused by random H02, C13 and O18 incoporation.

    Only O18, C13 and H02 attributions are considered so far.

    Parameters
    ----------
    label1 : str
        "No label" or formula e.g. "1C13 2H02"
    label2 : str
        "No label" or formula e.g. "1C13 2H02"
    res_corr_info : ResolutionCorrectionInfo
        Instance with resolution and mz calibration info.

    Returns
    -------
    float
        Transition probability of overlap
    """
    _logger.info(f"Indirect overlap prob for {label1} -> {label2}")
    # Label overlap possible with additional atoms
    coarse_mass_difference = calc_coarse_mass_difference(label1, label2)
    if (
        coarse_mass_difference <= 0
        or coarse_mass_difference >= res_corr_info.indirect_overlap_cutoff
    ):
        return 0

    probs = []
    for label_trans in generate_labels(coarse_mass_difference, res_corr_info):
        try:
            label1_mod = label1.add(label_trans)
        except ValueError as e:
            if "Too many atoms in label." in str(e):
                continue 
            else:
                raise e
        if is_isotologue_overlap(label1_mod, label2, res_corr_info):
            prob = calc_label_diff_prob(label1, label_trans)
            _logger.debug(f"For trans label {label_trans}: {prob}")
            probs.append(prob)
    prob_total = sum(probs)
    _logger.info(f"total indirect prob: {prob_total}\n")
    return prob_total


def generate_labels(mass_diff, res_corr_info):
    """Return combinations of possible isotopes e.g. H02 and C13 for given mass diff."""
    molecule_info = res_corr_info.molecule_info
    shift = molecule_info.isotopes.isotope_shift[molecule_info.get_isotopes()]
    isotopes = shift[shift > 0]  # only use isotope that cause mass shift
    for comb in itertools.product(range(0, mass_diff + 1), repeat=len(isotopes.index)):
        mass_shift = sum(isotopes * comb)
        if mass_shift != mass_diff:
            continue
        try:
            label = Label(pd.Series(comb, index=isotopes.index), molecule_info)
        except ValueError:
            continue
        yield label


def warn_direct_overlap(label_list, res_corr_info):
    """Warn if any of labels can have overlap."""
    for label1, label2 in itertools.permutations(label_list, 2):
        # Direct label overlap
        if is_isotologue_overlap(label1, label2, res_corr_info):
            warning = f"Direct overlap of {label1} and {label2}"
            warnings.warn(warning)
            _logger.warning(warning)



"""Unit tests for isotope correction."""

from pathlib import Path
import unittest

import pandas
from pandas.testing import assert_series_equal

import picor.isotope_probabilities as ip


__author__ = "Jørn Dietze"
__copyright__ = "Jørn Dietze"
__license__ = "gpl3"


class TestLabels(unittest.TestCase):
    """Label and formula parsing."""

    molecules_file = Path("tests/test_metabolites.csv")
    isotopes_file = Path("tests/test_isotopes.csv")
    molecule_info = ip.MoleculeInfo.get_molecule_info(
        molecule_name="Test1",
        molecules_file=molecules_file,
        isotopes_file=isotopes_file,
    )

    def test_init_label_dict(self):
        """Init Label instance with dict as input."""
        data = {"C13": 2, "O18": 3, "H02": 2}
        res = ip.Label(data, self.molecule_info)
        self.assertEqual(res.as_dict, data)

    def test_init_label_series(self):
        """Init Label instance with pandas Series as input."""
        data = pandas.Series([2, 3, 2], index=["C13", "O18", "H02"])
        res_corr = {"C13": 2, "O18": 3, "H02": 2}
        res = ip.Label(data, self.molecule_info)
        self.assertEqual(res.as_dict, res_corr)

    def test_eq_true(self):
        """Test __eq__ for same Label Instance."""
        data = {"C13": 2, "O18": 3, "H02": 2}
        label1 = ip.Label(data, self.molecule_info)
        label2 = ip.Label(data, self.molecule_info)
        self.assertTrue(label1 == label2)

    def test_eq_false(self):
        """Test __eq__ for different Label Instance."""
        data1 = {"C13": 2, "O18": 3, "H02": 2}
        data2 = {"C13": 1, "O18": 3, "H02": 2}
        label1 = ip.Label(data1, self.molecule_info)
        label2 = ip.Label(data2, self.molecule_info)
        self.assertFalse(label1 == label2)

    def test_eq_false_type(self):
        """Test __eq__ for differnt types."""
        data = {"C13": 2, "O18": 3, "H02": 2}
        label = ip.Label(data, self.molecule_info)
        iso = ip.IsotopeInfo(self.isotopes_file)
        self.assertFalse(label == iso)

    def test_lt(self):
        """Test __lt__."""
        label1 = ip.Label({"C13": 2, "H02": 3}, self.molecule_info)
        label2 = ip.Label({"O18": 3, "H02": 1}, self.molecule_info)
        self.assertTrue(label1 < label2)

    def test_ge(self):
        """Test __ge__."""
        label1 = ip.Label({"C13": 2, "H02": 3}, self.molecule_info)
        label2 = ip.Label({"O18": 1, "H02": 1}, self.molecule_info)
        label3 = ip.Label({"O18": 1, "H02": 3}, self.molecule_info)
        self.assertTrue(label1 >= label2)
        self.assertTrue(label1 >= label3)

    def test_le(self):
        """Test __le__."""
        label1 = ip.Label({"C13": 2, "H02": 3}, self.molecule_info)
        label2 = ip.Label({"O18": 3, "H02": 1}, self.molecule_info)
        label3 = ip.Label({"O18": 2, "H02": 1}, self.molecule_info)
        self.assertTrue(label1 <= label2)
        self.assertTrue(label1 <= label3)

    def test_gt(self):
        """Test __gt__."""
        label1 = ip.Label({"C13": 2, "H02": 3}, self.molecule_info)
        label2 = ip.Label({"O18": 1, "H02": 1}, self.molecule_info)
        self.assertTrue(label1 > label2)

    def test_formula_string(self):
        """Parse_formula parses string correctly."""
        data = "C30Si12H2NO1P2"
        res_corr = {"C": 30, "Si": 12, "H": 2, "N": 1, "O": 1, "P": 2}
        res = ip.parse_formula(data)
        self.assertEqual(res, res_corr)

    def test_formula_bad_type(self):
        """Parse_formula raises TypeError with list as input."""
        data = ["C12", "H12"]
        with self.assertRaises(TypeError):
            ip.parse_formula(data)

    def test_label_string(self):
        """Parse_label parses string correctly."""
        data = "2N153C13H02"
        res_corr = {"N15": 2, "C13": 3, "H02": 1}
        res = ip.Label(data, self.molecule_info)
        self.assertEqual(res.as_dict, res_corr)

    def test_label_wrong_atoms(self):
        """ValueError for atom not in molecule."""
        with self.assertRaises(ValueError):
            ip.Label("S33", self.molecule_info)

    def test_label_too_many_atoms(self):
        """ValueError for too many atoms in label."""
        with self.assertRaises(ValueError):
            ip.Label("55C13", self.molecule_info)

    def test_label_no_label(self):
        """Parse_label parses 'No label' string correctly."""
        data = "No label"
        res_corr = {}
        res = ip.Label(data, self.molecule_info)
        self.assertEqual(res.as_dict, res_corr)

    def test_label_bad_type(self):
        """Parse_label raises TypeError with list as input."""
        data = ["1C13", "H02"]
        with self.assertRaises(TypeError):
            ip.Label(data, self.molecule_info)

    def test_parse_label_zero_value(self):
        """Parse_label removes zero value elements."""
        string = "3C13 0N15 1H02"
        res_corr = {"C13": 3, "H02": 1}
        res = ip.Label.parse_label(string)
        self.assertEqual(res, res_corr)

    def test_parse_label_bad_type(self):
        """Parse_label raises TypeError for wrong Type."""
        bad_type_list = [["C13"], {"C13": 2}, 2]
        for data in bad_type_list:
            with self.assertRaises(TypeError):
                ip.Label.parse_label(data)

    def test_label_empty_string(self):
        """Parse_label raises ValueError with empty input string."""
        data = ""
        with self.assertRaises(ValueError):
            ip.Label(data, self.molecule_info)

    def test_label_split_label(self):
        """Parse_label parses composite label correctly."""
        data = "NAD:2N153C13H02"
        res_corr = {"N15": 2, "C13": 3, "H02": 1}
        res = ip.Label(data, self.molecule_info)
        self.assertEqual(res.as_dict, res_corr)

    def test_label_bad_split_label(self):
        """Parse_label raises ValueError with bad composite label."""
        data = "NAD:NamPT:2N153C13H02"
        with self.assertRaises(ValueError):
            ip.Label(data, self.molecule_info)

    def test_str_result(self):
        """Return correct string representation."""
        data = "2N153C13H02"
        res = str(ip.Label(data, self.molecule_info))
        res_corr = "Label: 2N153C13H02"
        self.assertEqual(res, res_corr)

    def test_generate_label_string(self):
        """generate_label_string return correct string."""
        data = {"N15": 2, "C13": 3, "H02": 1}
        res_corr = "2N15 3C13 1H02"
        res = ip.Label.generate_label_string(data)
        self.assertEqual(res, res_corr)

    def test_sort_list(self):
        """Sort_labels gives correct order with list of strings."""
        data = [
            ip.Label(label, self.molecule_info)
            for label in ["4C13", "3C13", "N154C13", "No label"]
        ]
        res_corr = ["No label", "3C13", "4C13", "N154C13"]
        res = ip.sort_labels(data)
        for t, c in zip(res, res_corr):
            self.assertEqual(t.as_string, c)

    def test_isotope_to_element_result(self):
        """Return correct elements."""
        label = ip.Label("2N15 3C13 1H02", self.molecule_info)
        res_corr = {"N": 2, "C": 3, "H": 1}
        res = label.isotope_to_element(label)
        self.assertEqual(res, res_corr)

    def test_add_result(self):
        """Return correct elements."""
        label1 = ip.Label("1N15 1C13", self.molecule_info)
        label2 = ip.Label("2N15 3C13 1H02", self.molecule_info)
        res_corr = ip.Label(
            pandas.Series({"C13": 4, "H02": 1, "N15": 3}), self.molecule_info
        )
        res = label2.add(label1)
        assert_series_equal(res.as_series, res_corr.as_series)

    def test_add_different_molecule(self):
        """Raise ValueError if labels have different molecule_info instances."""
        molecule_info1 = ip.MoleculeInfo.get_molecule_info(
            molecule_name="Test1",
            molecules_file=self.molecules_file,
            isotopes_file=self.isotopes_file,
        )
        molecule_info2 = ip.MoleculeInfo.get_molecule_info(
            molecule_name="Test4",
            molecules_file=self.molecules_file,
            isotopes_file=self.isotopes_file,
        )
        label1 = ip.Label("1N15 1C13", molecule_info1)
        label2 = ip.Label("2N15 3C13 1H02", molecule_info2)
        with self.assertRaises(ValueError):
            label1.add(label2)

    def test_subtract_result(self):
        """Return correct elements."""
        label1 = ip.Label("1N15 1C13", self.molecule_info)
        label2 = ip.Label("2N15 3C13 1H02", self.molecule_info)
        res_corr = ip.Label(
            pandas.Series({"C13": 2, "H02": 1, "N15": 1}), self.molecule_info
        )
        res = label2.subtract(label1)
        assert_series_equal(res.as_series, res_corr.as_series)

    def test_subtract_bad_type(self):
        """Raise TypeError for dict as label."""
        label1 = ip.Label("2O18 3C13 1H02", self.molecule_info)
        label2 = {"N": 1, "C": 2, "H": 1}
        with self.assertRaises(TypeError):
            label1.subtract(label2)

    def test_subtract_different_molecule(self):
        """Raise ValueError if labels have different molecule_info instances."""
        molecule_info1 = ip.MoleculeInfo.get_molecule_info(
            molecule_name="Test1",
            molecules_file=self.molecules_file,
            isotopes_file=self.isotopes_file,
        )
        molecule_info2 = ip.MoleculeInfo.get_molecule_info(
            molecule_name="Test4",
            molecules_file=self.molecules_file,
            isotopes_file=self.isotopes_file,
        )
        label1 = ip.Label("1N15 1C13", molecule_info1)
        label2 = ip.Label("2N15 3C13 1H02", molecule_info2)
        with self.assertRaises(ValueError):
            label1.subtract(label2)

    def test_check_isotopes_error(self):
        """ValueError is raised for unknown isotope."""
        with self.assertRaises(ValueError):
            ip.Label("2O15 3C13 1H02", self.molecule_info)

    def test_mass_nolabel(self):
        """Molecule mass calculation without label."""
        label = ip.Label("No label", self.molecule_info)
        res = label.calc_isotopologue_mass()
        self.assertAlmostEqual(res, 664.116947, places=5)

    def test_mass_label(self):
        """Molecule mass calculation with correct label."""
        label = ip.Label("15C13", self.molecule_info)
        res = label.calc_isotopologue_mass()
        self.assertAlmostEqual(res, 679.167270, places=5)


class TestLabelTuple(unittest.TestCase):
    """LabelTuple creation."""

    molecules_file = Path("tests/test_metabolites.csv")
    isotopes_file = Path("tests/test_isotopes.csv")
    molecule_info = ip.MoleculeInfo.get_molecule_info(
        molecule_name="Test1",
        molecules_file=molecules_file,
        isotopes_file=isotopes_file,
    )

    def test_init_result(self):
        """Correct results for list of labels."""
        data = ["C13", "2C13 2H02", "H02"]
        res = ip.LabelTuple(data, self.molecule_info)
        self.assertEqual(len(res), 3)
        self.assertEqual(res[2], ip.Label("2C13 2H02", self.molecule_info))
        self.assertEqual(res.molecule_info, self.molecule_info)

    def test_getitem(self):
        """Get value."""
        data = ["C13", "2C13 2H02", "H02"]
        lt = ip.LabelTuple(data, self.molecule_info)
        self.assertEqual(lt[2], ip.Label("2C13 2H02", self.molecule_info))


class TestIsotopeInfo(unittest.TestCase):
    """Isotope file parsing."""

    isotopes_file = Path("tests/test_isotopes.csv")

    def test_init_attributes_type(self):
        """Type of instance attributes."""
        instance = ip.IsotopeInfo(self.isotopes_file)
        self.assertIsInstance(instance.abundance, pandas.Series)
        self.assertIsInstance(instance.isotopes_file, Path)
        self.assertIsInstance(instance.isotope_mass_series, pandas.Series)

    def test_eq_true(self):
        """Test __eq__ for same IsotopeInfo Instance."""
        iso1 = ip.IsotopeInfo(self.isotopes_file)
        iso2 = ip.IsotopeInfo(self.isotopes_file)
        self.assertTrue(iso1 == iso2)

    def test_eq_false(self):
        """Test __eq__ for different IsotopeInfo Instance."""
        isotopes_file2 = Path("tests/test_isotopes2.csv")
        iso1 = ip.IsotopeInfo(self.isotopes_file)
        iso2 = ip.IsotopeInfo(isotopes_file2)
        self.assertFalse(iso1 == iso2)

    def test_eq_false_type(self):
        """Test __eq__ for differnt types."""
        iso = ip.IsotopeInfo(self.isotopes_file)
        other = {}
        self.assertFalse(iso == other)

    def test_get_isotope_abundance_result(self):
        """Return correct data."""
        res = ip.IsotopeInfo.get_isotope_abundance(self.isotopes_file)
        res_corr = [
            "H01",
            "H02",
            "C12",
            "C13",
            "N14",
            "N15",
            "O16",
            "O17",
            "O18",
            "Si28",
            "Si29",
            "Si30",
            "P31",
            "S32",
            "S33",
            "S34",
            "S35",
            "S36",
        ]
        self.assertListEqual(
            res_corr, list(res.index),
        )
        self.assertEqual(res["C12"], 0.9893)

    def test_get_isotope_mass_series_keys(self):
        """Return correct Series keys."""
        res = ip.IsotopeInfo.get_isotope_mass_series(self.isotopes_file)
        self.assertListEqual(
            ["H01", "H02", "C12"], list(res.keys())[:3],
        )

    def test_get_isotope_mass_series_fields(self):
        """Return correct data."""
        res = ip.IsotopeInfo.get_isotope_mass_series(self.isotopes_file)
        self.assertEqual(13.003354835, res["C13"])

    def test_get_isotope_mass_series_names(self):
        """Return correct index and data names."""
        res = ip.IsotopeInfo.get_isotope_mass_series(self.isotopes_file)
        self.assertEqual("isotope", res.keys().name)
        self.assertEqual("mass", res.name)

    def test_get_elements(self):
        """Return correct list of elements."""
        res = ip.IsotopeInfo.get_elements(self.isotopes_file)
        self.assertSetEqual({"H", "C", "O", "N", "Si", "S", "P"}, res)

    def test_get_isotopes_from_elements(self):
        """Correct list of isotopes."""
        iso = ip.IsotopeInfo(self.isotopes_file)
        ele = ["C", "O", "H"]
        res = iso.get_isotopes_from_elements(ele)
        res_corr = ["C12", "C13", "O16", "O17", "O18", "H01", "H02"]
        self.assertListEqual(res, res_corr)


class TestMoleculeInfo(unittest.TestCase):
    """Label and formula parsing."""

    molecules_file = Path("tests/test_metabolites.csv")
    molecules_file2 = Path("tests/test_metabolites_missing_charge.csv")
    isotopes_file = Path("tests/test_isotopes.csv")

    def test_init_formula(self):
        """Return right fields"""
        isotopes = ip.IsotopeInfo(self.isotopes_file)
        formula = {"C": 21, "H": 28, "N": 7, "O": 14, "P": 2}
        charge = 1
        ins = ip.MoleculeInfo(formula, charge, isotopes)
        self.assertIsInstance(ins, ip.MoleculeInfo)

    def test_get_molecule_info_formula_type(self):
        """Type of instance attributes."""
        ins = ip.MoleculeInfo.get_molecule_info(
            molecule_formula="C21H28N7O14P2",
            molecule_charge=1,
            isotopes_file=self.isotopes_file,
        )
        self.assertIsInstance(ins.isotopes, ip.IsotopeInfo)
        self.assertIsInstance(ins.charge, int)
        self.assertIsInstance(ins.formula, dict)

    def test_get_molecule_info_formula_result(self):
        """Return right fields"""
        ins = ip.MoleculeInfo.get_molecule_info(
            molecule_formula="C21H28N7O14P2",
            molecule_charge=1,
            isotopes_file=self.isotopes_file,
        )
        self.assertEqual(ins.charge, +1)
        self.assertDictEqual(ins.formula, {"C": 21, "H": 28, "N": 7, "O": 14, "P": 2})

    def test_get_molecule_info_name_type(self):
        """Type of instance attributes."""
        ins = ip.MoleculeInfo.get_molecule_info(
            molecule_name="Test1",
            molecules_file=self.molecules_file,
            isotopes_file=self.isotopes_file,
        )
        self.assertIsInstance(ins.isotopes, ip.IsotopeInfo)
        self.assertIsInstance(ins.charge, int)
        self.assertIsInstance(ins.formula, dict)

    def test_get_molecule_info_name_result(self):
        """Return right fields"""
        ins = ip.MoleculeInfo.get_molecule_info(
            molecule_name="Test1",
            molecules_file=self.molecules_file,
            isotopes_file=self.isotopes_file,
        )
        self.assertEqual(ins.charge, +1)
        self.assertDictEqual(ins.formula, {"C": 21, "H": 28, "N": 7, "O": 14, "P": 2})

    def test_get_molecule_info_wrong_arguments(self):
        """Raise ValueError when too many arguments."""
        with self.assertRaises(ValueError):
            ip.MoleculeInfo.get_molecule_info(
                molecule_name="Test1",
                molecule_formula="C21H28N7O14P2",
                molecules_file=self.molecules_file,
                isotopes_file=self.isotopes_file,
            )

    def test_eq_true(self):
        """Test __eq__ for same MoleculeInfo Instances."""
        molecule1 = ip.MoleculeInfo.get_molecule_info(
            molecule_name="Test1",
            molecules_file=self.molecules_file,
            isotopes_file=self.isotopes_file,
        )
        molecule2 = ip.MoleculeInfo.get_molecule_info(
            molecule_name="Test1",
            molecules_file=self.molecules_file,
            isotopes_file=self.isotopes_file,
        )
        self.assertTrue(molecule1 == molecule2)

    def test_eq_false(self):
        """Test __eq__ for different MoleculeInfo Instance."""
        molecule1 = ip.MoleculeInfo.get_molecule_info(
            molecule_name="Test1",
            molecules_file=self.molecules_file,
            isotopes_file=self.isotopes_file,
        )
        molecule2 = ip.MoleculeInfo.get_molecule_info(
            molecule_name="Test4",
            molecules_file=self.molecules_file,
            isotopes_file=self.isotopes_file,
        )
        self.assertFalse(molecule1 == molecule2)

    def test_eq_false_type(self):
        """Test __eq__ for different types."""
        molecule = ip.MoleculeInfo.get_molecule_info(
            molecule_name="Test1",
            molecules_file=self.molecules_file,
            isotopes_file=self.isotopes_file,
        )
        iso = ip.IsotopeInfo(self.isotopes_file)
        self.assertFalse(molecule == iso)

    def test_molecule_name_unknown(self):
        """Raise KeyError with undefined molecule name."""
        with self.assertRaises(KeyError):
            ip.MoleculeInfo.get_molecule_info(
                molecule_name="Unknown",
                molecules_file=self.molecules_file,
                isotopes_file=self.isotopes_file,
            )

    def test_get_formula_invalid_element(self):
        """Raise ValueError with invalid element in molecule formula."""
        with self.assertRaises(ValueError):
            ip.MoleculeInfo.get_molecule_info(
                molecule_name="Test3",
                molecules_file=self.molecules_file,
                isotopes_file=self.isotopes_file,
            )

    def test_get_charge_result(self):
        """Correct molecule charge."""
        res = ip.MoleculeInfo.get_charge("Test1", self.molecules_file)
        self.assertEqual(res, 1)

    def test_get_charge_column_missing(self):
        """ValueError if value in 'charge' column is missing."""
        with self.assertRaises(ValueError):
            ip.MoleculeInfo.get_charge("Test4", self.molecules_file2)

    def test_get_elements_result(self):
        """Return correct list of elements."""
        molecule = ip.MoleculeInfo.get_molecule_info(
            molecule_name="Test1",
            molecules_file=self.molecules_file,
            isotopes_file=self.isotopes_file,
        )
        res = molecule.get_elements()
        res_corr = ["C", "H", "N", "O", "P"]
        self.assertListEqual(res, res_corr)

    def test_get_isotopes_result(self):
        """Return correct list of isotopes."""
        molecule = ip.MoleculeInfo.get_molecule_info(
            molecule_name="Test1",
            molecules_file=self.molecules_file,
            isotopes_file=self.isotopes_file,
        )
        res = molecule.get_isotopes()
        res_corr = [
            "C12",
            "C13",
            "H01",
            "H02",
            "N14",
            "N15",
            "O16",
            "O17",
            "O18",
            "P31",
        ]
        self.assertListEqual(res, res_corr)


class TestCorrectionFactor(unittest.TestCase):
    """Calculation of correction factor."""

    molecules_file = Path("tests/test_metabolites.csv")
    isotopes_file = Path("tests/test_isotopes.csv")
    molecule_info = ip.MoleculeInfo.get_molecule_info(
        molecule_name="Test1",
        molecules_file=molecules_file,
        isotopes_file=isotopes_file,
    )

    def test_result_no_label(self):
        """Result without label."""
        res = ip.calc_correction_factor(self.molecule_info)
        self.assertAlmostEqual(res, 1.33471643)

    def test_result_with_label(self):
        """Result with complex label."""
        res = ip.calc_correction_factor(
            self.molecule_info, label=ip.Label("10C13 1N15 12H02", self.molecule_info),
        )
        self.assertAlmostEqual(res, 1.19257588)

    def test_result_wrong_label(self):
        """ValueError with impossible label."""
        with self.assertRaises(ValueError):
            ip.calc_correction_factor(
                self.molecule_info, label=ip.Label("100C131N15", self.molecule_info),
            )


class TestDiffProb(unittest.TestCase):
    """Calculation of difference probability."""

    molecules_file = Path("tests/test_metabolites.csv")
    isotopes_file = Path("tests/test_isotopes.csv")
    molecule_info = ip.MoleculeInfo.get_molecule_info(
        molecule_name="Test1",
        molecules_file=molecules_file,
        isotopes_file=isotopes_file,
    )

    def test_result_empty_diff_label(self):
        """Result with empty diff label."""
        label = ip.Label("1C13", self.molecule_info)
        diff_label = ip.Label({}, self.molecule_info)
        res = ip.calc_label_diff_prob(label, diff_label)
        self.assertEqual(res, 0)

    def test_result_negative_diff_label(self):
        """Result with negative diff label."""
        label = ip.Label("1C13", self.molecule_info)
        diff_label = ip.Label({"C13": -1, "N15": 1}, self.molecule_info)
        res = ip.calc_label_diff_prob(label, diff_label)
        self.assertEqual(res, 0)

    def test_result_simple_diff_label(self):
        """Result with simple one atom diff label."""
        label = ip.Label("1C13", self.molecule_info)
        diff_label = ip.Label("1C13", self.molecule_info)
        res = ip.calc_label_diff_prob(label, diff_label)
        self.assertAlmostEqual(res, 0.1744399)

    def test_result_complex_diff_label(self):
        """Result with diff label containing multiple atoms."""
        label = ip.Label("1C13", self.molecule_info)
        diff_label = ip.Label("1C13 1N15", self.molecule_info)
        res = ip.calc_label_diff_prob(label, diff_label)
        self.assertAlmostEqual(res, 0.0043485)



"""Isotopologue correction functions.

Functions:
    calc_correction_factor: Get correction factor for molecule and label.
    calc_transition_prob: Get transition probablity for two isotopologues.
"""
from collections.abc import Sequence
from functools import reduce
from operator import mul
import re

import pandas as pd
from scipy.special import binom

__author__ = "Jørn Dietze"
__copyright__ = "Jørn Dietze"
__license__ = "gpl3"


class IsotopeInfo:
    """Class containing isotope data."""

    def __init__(self, isotopes_file):
        """Data class containing isotope information.

        Parameter
        ---------
        isotopes_file : str or Path
            File path to isotopes file
            comma separated csv with element and abundance columns
        """
        self.isotopes_file = isotopes_file
        self.abundance = self.get_isotope_abundance(isotopes_file)
        self.isotope_mass_series = self.get_isotope_mass_series(isotopes_file)
        self.isotope_shift = self.get_shift(isotopes_file)
        self.elements = self.get_elements(isotopes_file)

    def __repr__(self):
        return f"IsotopeInfo('{self.isotopes_file}')"

    def __eq__(self, other):
        if isinstance(other, IsotopeInfo):
            return self.isotopes_file == other.isotopes_file
        return False

    @staticmethod
    def get_isotope_abundance(isotopes_file):
        """Get abundace of different isotopes.

        Parse file with abundance of different isotopes

        Parameters
        ----------
        isotopes_file : str or Path
            File path to isotopes file
            comma separated csv with element and abundance columns

        Returns
        -------
        dict
            Abundance of isotope.
        """
        abundance = pd.read_csv(
            isotopes_file,
            sep="\t",
            usecols=["abundance", "isotope"],
            index_col="isotope",
        ).squeeze("columns")

        return abundance

    @staticmethod
    def get_isotope_mass_series(isotopes_file):
        """Get series of isotope masses.

        Parameters
        ----------
        isotopes_file : str or Path
            File path to isotopes file
            comma separated csv with element and abundance columns

        Returns
        -------
        pandas.Series
            Isotope names and masses
        """
        return pd.read_csv(
            isotopes_file,
            sep="\t",
            usecols=["mass", "isotope"],
            index_col="isotope",
        ).squeeze("columns")

    @staticmethod
    def get_shift(isotopes_file):
        """Get series of mass shift between isotopes."""
        iso_data = pd.read_csv(
            isotopes_file,
            sep="\t",
            index_col="isotope",
        )
        iso_data = iso_data.groupby("element", group_keys=False).apply(
            lambda gdf: gdf.assign(
                shift=lambda df: round(df["mass"] - min(df["mass"]))
            ),
            include_groups=False,
        )
        return iso_data["shift"].astype("int64")

    @staticmethod
    def get_elements(isotopes_file):
        """Get set of elements in isotopes_file."""
        elements = pd.read_csv(
            isotopes_file,
            sep="\t",
            usecols=["element"],
        ).squeeze("columns")
        return set(elements)

    def get_isotopes_from_elements(self, element_list):
        """Get list of isotopes based on elements."""
        iso_data = pd.read_csv(
            self.isotopes_file,
            sep="\t",
            index_col="element",
            usecols=["element", "isotope"],
        ).squeeze("columns")

        return list(iso_data[element_list])

    def get_isotopes(self):
        """Get list of all isotopes."""
        return list(self.isotope_mass_series.keys())

    @staticmethod
    def get_element_from_isotope(isotope):
        """Return element based on isotope."""
        return re.match(r"[A-Z][a-z]?", isotope).group()

    def get_lightest_isotope_from_element(self, element):
        """Return lightest isotpe based on element, e.g. 'C12' from 'C'."""
        iso_data = pd.read_csv(
            self.isotopes_file,
            sep="\t",
            index_col="element",
            usecols=["element", "isotope"],
        ).squeeze("columns")
        iso_data = iso_data.groupby("element", group_keys=False).min()
        return iso_data[element]

    def get_lightest_isotope_from_isotope(self, isotope):
        """Return lightest isotope of element for given isotope."""
        return self.get_lightest_isotope_from_element(
            self.get_element_from_isotope(isotope)
        )


class MoleculeInfo:
    """Class containing molecule data."""

    def __init__(self, formula, charge, isotopes):
        """Class contains molecule formula, charge and isotope data.

        Parameters
        ----------
        formula : dict
            Elements as keys
            Number of atoms as values
            E.g. {"C":3, "H":8}
        charge : int
            Charge as signed integer
        isotopes : IsotopeInfo
            Contains isotope data.
        """
        self.isotopes = isotopes
        self.formula = formula
        self.charge = int(charge)

    def __repr__(self):
        return f"MoleculeInfo({self.formula}, {self.charge}, {self.isotopes})"

    def __eq__(self, other):
        if isinstance(other, MoleculeInfo):
            return (
                self.formula == other.formula
                and self.charge == other.charge
                and self.isotopes == other.isotopes
            )
        return False

    @classmethod
    def get_molecule_info(
        cls,
        molecule_name=None,
        molecules_file=None,
        molecule_formula=None,
        molecule_charge=None,
        isotopes_file=None,
    ):
        """Get MoleculeInfo instance from either formula or name.

        Either name and molecules file or formula and charge are reqired.

        Parameters
        ----------
        molecule_name : str
            Molecule name as in molecules_file.
        molecules_file : str or Path
            tab-separated file with name, formula and charge as rows
            e.g. Suc C4H4O3 -1
        molecule_formula : str
            Chemical formula as string.
            No spaces or underscores allowed.
            E.g. "C3H7O1"
        molecule_charge : int
            Charge as signed integer
        isotopes_file : str or Path
            File path to isotopes file
            comma separated csv with element and abundance columns

        Raises
        ------
        ValueError
            If both molecule_name and molecule_formula are specified.

        Returns
        -------
        MoleculeInfo
            Instance of class with formula, charge and isotope data.
        """
        if molecule_name and molecules_file and not molecule_formula:
            molecule = cls.create_from_name(
                molecule_name, molecules_file, isotopes_file
            )
        elif molecule_formula and molecule_charge and not molecule_name:
            molecule = cls.create_from_formula(
                molecule_formula, molecule_charge, isotopes_file
            )
        else:
            raise ValueError(
                "Either molecule name and file or molecule formula and charge have to be specified"
            )
        return molecule

    @classmethod
    def create_from_name(cls, molecule_name, molecules_file, isotopes_file):
        """Get MoleculeInfo instance from molecule name and file."""
        isotopes = IsotopeInfo(isotopes_file)
        molecule_list = cls.get_molecule_list(molecules_file)
        molecule_formula = cls.get_formula(molecule_name, molecule_list, isotopes)
        molecule_charge = cls.get_charge(molecule_name, molecules_file)
        return cls(molecule_formula, molecule_charge, isotopes)

    @classmethod
    def create_from_formula(cls, molecule_formula, molecule_charge, isotopes_file):
        """Get MoleculeInfo instance from formula and charge."""
        isotopes = IsotopeInfo(isotopes_file)
        formula = parse_formula(molecule_formula)
        return cls(formula, molecule_charge, isotopes)

    @staticmethod
    def get_molecule_list(molecules_file):
        """Read and return molecule_list."""
        molecule_list = pd.read_csv(molecules_file, sep="\t", na_filter=False)
        molecule_list["formula"] = molecule_list["formula"].apply(parse_formula)
        molecule_list.set_index("name", drop=True, inplace=True)
        return molecule_list

    @staticmethod
    def get_formula(molecule_name, molecule_list, isotopes):
        """Return molecular formula.

        Parse and look up molecular formula of molecule and
        return number of atoms per element

        Returns
        -------
        dict
            Elements as keys
            Number of atoms as values

        Raises
        ------
        KeyError
            If molecule is not contained in molecules_file.
        ValueError
            If unknown element in molecule formula.
        """
        try:
            n_atoms = molecule_list.loc[molecule_name].formula
        except KeyError as error:
            raise KeyError(
                f"Molecule {error} couldn't be found in molecules file"
            ) from error
        if not all(element in isotopes.elements for element in n_atoms):
            raise ValueError("Unknown element in molecule")
        return n_atoms

    @staticmethod
    def get_charge(molecule_name, molecules_file):
        """Get charge of molecule."""
        charges = pd.read_csv(
            molecules_file,
            sep="\t",
            usecols=["name", "charge"],
            index_col="name",
        ).squeeze("columns")
        try:
            charges = charges.astype(int)
        except ValueError as exc:
            raise ValueError(
                "Charge of at least one molecule missing in molecules_file or missformed file"
            ) from exc
        return charges[molecule_name]

    def get_elements(self):
        """Return elements of molecules."""
        return list(self.formula.keys())

    def get_isotopes(self):
        """Return all possible isotopes in molecule."""
        elements = self.get_elements()
        return self.isotopes.get_isotopes_from_elements(elements)

    def get_molecule_light_isotopes(self):
        """Replace all element names with light isotopes ("C" -> "C13")."""
        molecule_series = pd.Series(
            self.formula,
            dtype="int64",
        )
        result = molecule_series.rename(
            {
                "H": "H01",
                "C": "C12",
                "N": "N14",
                "O": "O16",
                "Si": "Si28",
                "P": "P31",
                "S": "S32",
            }
        )
        return result

    @staticmethod
    def subtract_label(molecule_series, label):
        """Subtract label atoms from molecule formula.

        Parameters
        ----------
        molecule_series : Series
            Formula as Series with isotopes as index
        label : Label
            Type of isotopic label, e.g. Label("1N15")

        Returns
        -------
        Series
            Formula with isotopes as index

        Raises
        ------
        ValueError
            If too many atoms are subtracted and negative number of atoms is reached
        """
        formula_difference = molecule_series.copy()
        for heavy in label.as_series.keys():
            light = label.molecule_info.isotopes.get_lightest_isotope_from_isotope(
                heavy
            )
            formula_difference[light] = molecule_series[light] - label.as_series[heavy]
        return formula_difference


class Label:
    """Class for storing label information."""

    def __init__(self, label, molecule_info):
        """Class contains label in different repressentations (dict, series, str).

        Parameters
        ----------
        label : str, dict, Series
            Either string: "No label" or label formula, e.g. "2C13 3H02"
            or dict: {} or e.g. {"C13": 2, "H02": 3}
            or pandas Series: Series() or e.g. Series([2, 3], index=["C13", "H02"])
        molecule_info : MoleculeInfo
            Instance with molecule and isotope information.

        Raises
        ------
        TypeError
            If label not str, dict or pandas Series
        """
        # TODO Use OrderedDict to have unique solution
        if isinstance(label, str):
            self.as_string = label
            self.as_dict = self.parse_label(label)
            self.as_series = pd.Series(self.as_dict, dtype="int64")
        elif isinstance(label, dict):
            self.as_dict = {ele: n for ele, n in label.items() if n}
            self.as_series = pd.Series(self.as_dict, dtype="int64")
            self.as_string = self.generate_label_string(self.as_dict)
        elif isinstance(label, pd.Series):
            self.as_series = label[label != 0].astype("int64")
            self.as_dict = dict(self.as_series)
            self.as_string = self.generate_label_string(self.as_dict)
        else:
            raise TypeError("label has to be str, dict or pandas Series")
        self.molecule_info = molecule_info
        self.mass = self.get_coarse_mass_shift()  # Coarse mass of label alone
        self.check_isotopes()
        self.check_number_atoms()

    def __repr__(self):
        return f"Label({self.as_dict}, {self.molecule_info})"

    def __str__(self):
        return f"Label: {self.as_string}"

    def __bool__(self):
        return bool(self.as_dict)

    def __eq__(self, other):
        if isinstance(other, Label):
            return (
                self.as_dict == other.as_dict
                and self.molecule_info == other.molecule_info
            )
        return False

    def __gt__(self, other):
        return self.mass > other.mass

    def __lt__(self, other):
        return self.mass < other.mass

    def __ge__(self, other):
        return self.mass >= other.mass

    def __le__(self, other):
        return self.mass <= other.mass

    def check_isotopes(self):
        """Raise ValueError for label isotopes not in isotope list."""
        for iso in self.as_dict.keys():
            if iso not in self.molecule_info.isotopes.get_isotopes():
                raise ValueError(f"Label atom {iso} not in isotopes file")

    def check_number_atoms(self):
        """Raise ValueError if number of atoms in label is greater than molecule."""
        for iso, num in self.as_dict.items():
            elem = self.molecule_info.isotopes.get_element_from_isotope(iso)
            try:
                if num > self.molecule_info.formula[elem]:
                    raise ValueError("Too many atoms in label.")
            except KeyError as err:
                raise ValueError("Label contains atoms not in molecule.") from err

    def subtract(self, other):
        """Return difference between two labels as new Label instance.
        self - other
        """
        if not isinstance(other, Label):
            raise TypeError("Both labels has to be Label instance")
        if self.molecule_info != other.molecule_info:
            raise ValueError("Labels have different molecule_info.")
        return Label(
            self.as_series.sub(other.as_series, fill_value=0).astype("int64"),
            self.molecule_info,
        )

    def add(self, other):
        """Add other label and return new Label instance.

        Parameters
        ----------
        other : Label

        Returns
        -------
        Label
            New summed label

        Raises
        ------
        ValueError
            Both labels have different molecule information.
        """
        if self.molecule_info != other.molecule_info:
            raise ValueError("Labels have different molecule_info.")
        return Label(
            self.as_series.add(other.as_series, fill_value=0), self.molecule_info
        )

    def get_coarse_mass_shift(self):
        """Return mass of label compared to unlabelled molecule.
        E.g. 5 for "2C13 2H02"
        """
        iso_shift = self.molecule_info.isotopes.isotope_shift
        return int(self.as_series.multiply(iso_shift).sum())

    def calc_isotopologue_mass(self):
        """Calculate mass of isotopologue.

        Given the molecule name and label composition, return mass in atomic units.

        Returns
        -------
        float
            Molecule mass
        """
        molecule_series = self.molecule_info.get_molecule_light_isotopes()
        light_isotopes = self.molecule_info.subtract_label(molecule_series, self)
        formula_isotopes = pd.concat([light_isotopes, self.as_series])
        mass = (
            self.molecule_info.isotopes.isotope_mass_series.multiply(formula_isotopes)
            .dropna()
            .sum()
        )
        return float(mass)

    @staticmethod
    def generate_label_string(label_as_dict, sep=" "):
        """Generate string representation of label instance."""
        return sep.join([f"{n_atom}{iso}" for iso, n_atom in label_as_dict.items()])

    @staticmethod
    def parse_label(string):
        """Parse label e.g. "3C13 2H02"  and return dict.

        Parse label e.g. 5C13N15 and return dictionary of elements and number of atoms
        Prefix separated by colon can be used and will be ignored, e.g. "NA:2C13"
        Underscores can be used to separate different elements, e.g. "3C13_6H02"

        Parameters
        ----------
        string : str
            Chemical formula

        Returns
        -------
        dict
            Elements as keys
            Number as values

        Raises
        ------
        TypeError
            If input is not string.
        ValueError
           If label contains more than one colon.
        ValueError
            If label string is empty.
        """
        if not isinstance(string, str):
            raise TypeError("label must be string")
        if string.count(":") > 1:
            raise ValueError("only one colon allowed in label to separate prefix")
        string = string.split(":")[-1]
        if string == "":
            raise ValueError(
                "Empty string as label not allowed.\n"
                "Use 'No label' instead or specify label."
            )

        if string.lower() == "no label":
            return {}
        label = re.findall(r"(\d*)([A-Z][a-z]?\d\d)", string)
        label_dict = {}
        for num, elem in label:
            if num == "":
                label_dict[elem] = 1
            elif int(num) > 0:
                label_dict[elem] = int(num)
        return label_dict

    @staticmethod
    def isotope_to_element(label):
        """Change dict key from isotope to element (e.g. H02 -> H)."""
        return {
            re.match(r"[A-Z][a-z]?", elem).group(): n
            for elem, n in label.as_dict.items()
        }


class LabelTuple(Sequence):
    """Class to manage multiple labels."""

    __slots__ = ("_items", "molecule_info")

    def __init__(self, label_list, molecule_info):
        """Class contains tuple of labels from list of strings.

        Parameters
        ----------
        label_list : list
            List of label strings
        molecule_info : MoleculeInfo
            Instance with molecule and isotope information.
        """
        labels = [Label(label, molecule_info) for label in label_list]
        self._items = sort_labels(labels)
        self.molecule_info = molecule_info

    def __getitem__(self, index):
        return self._items[index]

    def __len__(self):
        return len(self._items)

    def __repr__(self):
        return f"LabelTuple({[label.as_string for label in self._items]}, {self.molecule_info})"


def parse_formula(string):
    """Parse chemical formula and return dict.

    Parse chemical formula of type C3O3H6 and return dictionary
    of elements and number of atoms.
    Be careful, no input check is happening!

    Parameters
    ----------
    string : str
        chemical formula as string

    Returns
    -------
    dict
        Elements as keys
        Number as values
    """
    elements = re.findall(r"([A-Z][a-z]?)(\d*)", string)
    return {elem: int(num) if num != "" else 1 for elem, num in elements}


def sort_labels(labels):
    """Sort list of molecule labels by coarse mass.

    Sort list of molecule labels by coarse mass
    (number of neutrons) from lower to higher mass

    Parameters
    ----------
    labels : list of Labels
        All elements of list must be of instance Label

    Returns
    -------
    list of Labels
        Sorted list
    """
    masses = [(label, label.mass) for label in labels]
    sorted_masses = sorted(masses, key=lambda tup: tup[1])
    sorted_labels = [lab[0] for lab in sorted_masses]
    return sorted_labels


def calc_correction_factor(
    molecule_info,
    label=False,
):
    """Calculate correction factor with molecule composition defined by label.

    Label supports only H02 (deuterium), C13, N15 and 'No label'.

    Parameters
    ----------
    molecule_info : MoleculeInfo
        Instance with molecule and isotope information.
    label : False or Label
        Instance of Label
        Type and number of atoms.
        E.g. Label('5C13N15') or Label('No label'). False equals to no label

    Returns
    -------
    float
        Correction factor

    Raises
    ------
    ValueError
        If there's more labelled atoms than atoms.
    """
    atom_label = Label.isotope_to_element(label) if label else {}

    n_atoms = molecule_info.formula
    abundance = molecule_info.isotopes.abundance
    prob = {}
    for elem in n_atoms:
        n_atom = (
            n_atoms[elem] - atom_label[elem] if elem in atom_label else n_atoms[elem]
        )
        isotope = molecule_info.isotopes.get_lightest_isotope_from_element(elem)
        prob[elem] = abundance[isotope] ** n_atom
    probability = reduce(mul, prob.values())
    return 1 / probability


def calc_label_diff_prob(label, difference_label):
    """Calculate the transition probablity of difference in labelled atoms.

    Parameters
    ----------
    label : Label
        Type of isotopic label, e.g. Label("1N15")
    difference_label : Label
        Type of isotopic label, e.g. Label("1N15")

    Returns
    -------
    float
        Transition probability
    """
    if not difference_label:
        return 0

    # No classical transition possible with negative values
    if difference_label.as_series.lt(0).any():
        return 0
    n_atoms = label.molecule_info.formula
    abundance = label.molecule_info.isotopes.abundance

    prob = []
    for isotope_lab, n_diff_label in difference_label.as_series.items():
        # breakpoint()
        # get number of atoms of isotope, default to 0
        elem = label.molecule_info.isotopes.get_element_from_isotope(isotope_lab)
        isotope_unlab = label.molecule_info.isotopes.get_lightest_isotope_from_isotope(
            isotope_lab
        )

        n_label = label.as_dict.get(isotope_lab, 0)
        n_unlab = n_atoms[elem] - n_label - n_diff_label
        abun_lab = abundance[isotope_lab]
        abun_unlab = abundance[isotope_unlab]

        trans_pr = binom((n_atoms[elem] - n_label), n_diff_label)
        trans_pr *= abun_lab**n_diff_label
        trans_pr *= abun_unlab**n_unlab
        prob.append(trans_pr)

    # Prob is product of single probabilities
    prob_total = reduce(mul, prob)
    assert prob_total <= 1, "Transition probability greater than 1"
    return prob_total



"""Unit tests for isotope correction."""

from pathlib import Path
import unittest

import pandas as pd

import picor.isotope_correction as ic
import picor.isotope_probabilities as ip
import picor.resolution_correction as rc

__author__ = "Jørn Dietze"
__copyright__ = "Jørn Dietze"
__license__ = "gpl3"


class TestIsotopologueCorrection(unittest.TestCase):
    """Total Correction Factor for molecule and DataFrame."""

    molecules_file = Path("tests/test_metabolites.csv")
    isotopes_file = Path("tests/test_isotopes.csv")

    def test_result(self):
        """Result with default values."""
        data = pd.read_csv(Path("tests/test_dataset.csv"), index_col=0)
        data.drop(columns=["dummy column int", "dummy column str"], inplace=True)
        molecule_name = "NAD"
        res = ic.calc_isotopologue_correction(data, molecule_name,)
        data_corrected = pd.read_csv(
            Path("tests/test_dataset_corrected.csv"), index_col=0
        )
        pd.testing.assert_frame_equal(data_corrected, res)

    def test_result_files(self):
        """Result with given molecule and isotopes file paths."""
        data = pd.read_csv(Path("tests/test_dataset.csv"), index_col=0)
        data.drop(columns=["dummy column int", "dummy column str"], inplace=True)
        molecule_name = "Test1"
        res = ic.calc_isotopologue_correction(
            data,
            molecule_name,
            molecules_file=self.molecules_file,
            isotopes_file=self.isotopes_file,
        )
        data_corrected = pd.read_csv(
            Path("tests/test_dataset_corrected.csv"), index_col=0
        )
        pd.testing.assert_frame_equal(data_corrected, res)

    def test_result_subset(self):
        """Result with explicit subset."""
        data = pd.read_csv(Path("tests/test_dataset.csv"), index_col=0)
        subset = ["No label", "1C13", "4C13 6H02 3N15"]
        molecule_name = "Test1"
        res = ic.calc_isotopologue_correction(
            data,
            molecule_name,
            subset=subset,
            molecules_file=self.molecules_file,
            isotopes_file=self.isotopes_file,
        )
        assert res.shape == data.shape
        res.drop(columns=["dummy column int", "dummy column str"], inplace=True)
        data_corrected = pd.read_csv(
            Path("tests/test_dataset_corrected.csv"), index_col=0
        )
        pd.testing.assert_frame_equal(data_corrected, res)

    def test_result_exclude_col(self):
        """Result with exclude_col parameter."""
        data = pd.read_csv(Path("tests/test_dataset.csv"), index_col=0)
        exclude_col = ["dummy column int", "dummy column str"]
        molecule_name = "Test1"
        res = ic.calc_isotopologue_correction(
            data,
            molecule_name,
            exclude_col=exclude_col,
            molecules_file=self.molecules_file,
            isotopes_file=self.isotopes_file,
        )
        assert res.shape == data.shape
        res.drop(columns=["dummy column int", "dummy column str"], inplace=True)
        data_corrected = pd.read_csv(
            Path("tests/test_dataset_corrected.csv"), index_col=0
        )
        pd.testing.assert_frame_equal(data_corrected, res)

    def test_resolution_result(self):
        """Result with default values"""
        data = pd.read_csv(Path("tests/test_dataset_resolution.csv"), index_col=0)
        data.drop(columns=["dummy column int", "dummy column str"], inplace=True)
        molecule_name = "Test1"
        res = ic.calc_isotopologue_correction(
            data,
            molecule_name,
            resolution_correction=True,
            molecules_file=self.molecules_file,
            isotopes_file=self.isotopes_file,
        )
        data_corrected = pd.read_csv(
            Path("tests/test_dataset_resolution_corrected.csv"), index_col=0
        )
        pd.testing.assert_frame_equal(data_corrected, res)

    def test_resolution_ultra_high_result(self):
        """Result with ultra high resolution."""
        data = pd.read_csv(Path("tests/test_dataset.csv"), index_col=0)
        data.drop(columns=["dummy column int", "dummy column str"], inplace=True)
        data.rename(columns={"4C13 6H02 3N15": "2H02"}, inplace=True)
        molecule_name = "Test1"
        res_w_cor = ic.calc_isotopologue_correction(
            data,
            molecule_name,
            resolution_correction=True,
            resolution=1e8,
            molecules_file=self.molecules_file,
            isotopes_file=self.isotopes_file,
        )
        res_wo_cor = ic.calc_isotopologue_correction(
            data,
            molecule_name,
            resolution_correction=False,
            molecules_file=self.molecules_file,
            isotopes_file=self.isotopes_file,
        )
        pd.testing.assert_frame_equal(res_wo_cor, res_w_cor, rtol=1e-4)

    def test_resolution_warning(self):
        """Warn when overlapping isotopologues."""
        data = pd.read_csv(Path("tests/test_dataset.csv"), index_col=0)
        data.drop(columns=["dummy column int", "dummy column str"], inplace=True)
        data.rename(columns={"4C13 6H02 3N15": "1H02"}, inplace=True)
        molecule_name = "Test1"
        with self.assertWarns(UserWarning):
            ic.calc_isotopologue_correction(
                data,
                molecule_name,
                resolution_correction=True,
                molecules_file=self.molecules_file,
                isotopes_file=self.isotopes_file,
            )


class TestTransitionProbability(unittest.TestCase):
    """Calculation of probability between to isotopologues."""

    molecules_file = Path("tests/test_metabolites.csv")
    isotopes_file = Path("tests/test_isotopes.csv")
    molecule_info = ip.MoleculeInfo.get_molecule_info(
        molecule_name="Test4",
        molecules_file=molecules_file,
        isotopes_file=isotopes_file,
    )
    molecule_info2 = ip.MoleculeInfo.get_molecule_info(
        molecule_name="Test1",
        molecules_file=molecules_file,
        isotopes_file=isotopes_file,
    )
    res_corr_info = rc.ResolutionCorrectionInfo(False, 60000, 200, molecule_info)

    def test_result_label1_smaller(self):
        """Result with label1 being smaller than label2."""
        label1 = ip.Label("2N15", self.molecule_info)
        label2 = ip.Label("2N152C13", self.molecule_info)
        res = ic.calc_transition_prob(label1, label2, self.res_corr_info)
        self.assertAlmostEqual(res, 0.03685030)

    def test_result_label1_equal(self):
        """Result with label1 being equal to label2."""
        label1 = ip.Label("1N15", self.molecule_info)
        res = ic.calc_transition_prob(label1, label1, self.res_corr_info)
        self.assertEqual(res, 0)

    def test_result_label_missmatch_without_res_corr(self):
        """Zero result with no possible transition without res_corr."""
        label1 = ip.Label("2N15 4C13", self.molecule_info2)
        label2 = ip.Label("1N15 6C13", self.molecule_info2)
        res_corr_info = rc.ResolutionCorrectionInfo(
            False, 60000, 200, self.molecule_info2
        )
        res = ic.calc_transition_prob(label1, label2, res_corr_info)
        self.assertEqual(res, 0)

    def test_result_label_missmatch_with_res_corr(self):
        """Possible result for missmatch with resolution correction."""
        label1 = ip.Label("2N15 4C13", self.molecule_info2)
        label2 = ip.Label("1N15 6C13", self.molecule_info2)
        res_corr_info = rc.ResolutionCorrectionInfo(
            True, 60000, 200, self.molecule_info2
        )
        res = ic.calc_transition_prob(label1, label2, res_corr_info)
        self.assertAlmostEqual(res, 0.1794381)

    def test_result_label1_larger(self):
        """Result with label1 being larger than label2."""
        label1 = ip.Label("2N152C13", self.molecule_info)
        label2 = ip.Label("2N15", self.molecule_info)
        res = ic.calc_transition_prob(label1, label2, self.res_corr_info)
        self.assertEqual(res, 0)

    def test_result_molecule_formula(self):
        """Result with molecule formula."""
        molecule_info = ip.MoleculeInfo.get_molecule_info(
            molecule_name="Test1",
            molecules_file=self.molecules_file,
            isotopes_file=self.isotopes_file,
        )
        label1 = ip.Label("1N15", molecule_info)
        label2 = ip.Label("2N152C13", molecule_info)
        res = ic.calc_transition_prob(label1, label2, self.res_corr_info)
        self.assertAlmostEqual(res, 0.00042029)

    def test_wrong_type(self):
        """Type error with dict as molecule."""
        label1 = ip.Label("1N15", self.molecule_info)
        label2 = ip.Label("2N152C13", self.molecule_info)
        molecule = {"C": 12, "N": 15}
        with self.assertRaises(TypeError):
            ic.calc_transition_prob(label1, label2, molecule, self.res_corr_info)



"""Unit tests for isotope correction."""

import subprocess
from subprocess import PIPE


import pandas as pd

__author__ = "Jørn Dietze"
__copyright__ = "Jørn Dietze"
__license__ = "gpl3"


def test_result_csv(tmp_path):
    """Output file result with csv input."""
    outfile = tmp_path / "test_out.csv"
    res_correct = "./tests/test_dataset_corrected.csv"
    cmd = f"picor tests/test_dataset.csv NAD -x 'dummy column int' -x 'dummy column str' -o {outfile}"
    process = subprocess.run(cmd, shell=True)
    df_res = pd.read_csv(outfile, index_col=0)
    df_res.drop(columns=["dummy column str", "dummy column int"], inplace=True)
    df_corr = pd.read_csv(res_correct, index_col=0)
    assert process.returncode == 0
    pd.testing.assert_frame_equal(df_res, df_corr)


def test_result_xlsx(tmp_path):
    """Output file result with xlsx input."""
    outfile = tmp_path / "test_out.csv"
    res_correct = "./tests/test_dataset_corrected.csv"
    cmd = f"picor tests/test_dataset.xlsx NAD -x 'dummy column int' -x 'dummy column str' -o {outfile}"
    process = subprocess.run(cmd, shell=True)
    df_res = pd.read_csv(outfile, index_col=0)
    df_res.drop(columns=["dummy column str", "dummy column int"], inplace=True)
    df_corr = pd.read_csv(res_correct, index_col=0)
    assert process.returncode == 0
    pd.testing.assert_frame_equal(df_res, df_corr)


def test_result_wrong_fileformat(tmp_path):
    """Check error with unknown input."""
    cmd = "picor tests/test_dataset.tsv NAD -x 'dummy column int' -x 'dummy column str'"
    process = subprocess.run(cmd, shell=True, stdout=PIPE, stderr=PIPE)
    assert process.returncode == 1
    assert not process.stdout, "No output to stdout"
    assert len(process.stderr) > 0, "Error message printed to StdErr"


def test_result_resolution_correction(tmp_path):
    """Output file result with xlsx input."""
    outfile = tmp_path / "test_out.csv"
    res_correct = "./tests/test_dataset_resolution_corrected.csv"
    cmd = f"picor tests/test_dataset_resolution.csv NAD --res-correction -x 'dummy column int' -x 'dummy column str' -o {outfile}"
    subprocess.run(cmd, shell=True, check=True)
    df_res = pd.read_csv(outfile, index_col=0)
    df_res.drop(columns=["dummy column str", "dummy column int"], inplace=True)
    df_corr = pd.read_csv(res_correct, index_col=0)
    pd.testing.assert_frame_equal(df_res, df_corr)


def test_no_output(tmp_path):
    """Output on stdout."""
    cmd = "picor tests/test_dataset.csv NAD -x 'dummy column int' -x 'dummy column str'"
    process = subprocess.run(cmd, shell=True, stdout=PIPE, stderr=PIPE)
    assert process.returncode == 0
    assert not process.stderr, "No output to stderr"
    assert len(process.stdout) > 0, "Results not printed to stdout"


def test_logging_level_info(tmp_path):
    """Logging to stderr with INFO level."""
    cmd = "picor tests/test_dataset.csv NAD -x 'dummy column int' -x 'dummy column str' -v"
    process = subprocess.run(cmd, shell=True, check=True, stdout=PIPE, stderr=PIPE)
    log = process.stderr.decode("utf-8")
    assert len(process.stderr) > 0, "Logging not printed to stderr"
    assert len(process.stdout) > 0, "Results not printed to stdout"
    assert "INFO:picor.isotope_correction:" in log, "INFO not in stderr"


def test_logging_level_debug(tmp_path):
    """Logging to stderr with DEBUG level."""
    cmd = "picor tests/test_dataset.csv NAD -x 'dummy column int' -x 'dummy column str' -r -v -v"
    process = subprocess.run(cmd, shell=True, check=True, stdout=PIPE, stderr=PIPE)
    log = process.stderr.decode("utf-8")
    assert len(process.stderr) > 0, "Logging not printed to stderr"
    assert len(process.stdout) > 0, "Results not printed to stdout"
    assert "DEBUG:picor.resolution_correction:" in log, "DEBUG not in stderr"



"""Isotopologue correction using statistical distributions.

Functions:
    calc_isotopologue_correction: Correct DataFrame with measurements.
"""
import logging
import os
from pathlib import Path

import picor.isotope_probabilities as ip
import picor.resolution_correction as rc

__author__ = "Jørn Dietze"
__copyright__ = "Jørn Dietze"
__license__ = "gpl3"

_logger = logging.getLogger(__name__)


def calc_isotopologue_correction(
    raw_data,
    molecule_name=None,
    molecules_file=None,
    molecule_formula=None,
    molecule_charge=None,
    subset=False,
    exclude_col=False,
    resolution_correction=False,
    mz_calibration=200,
    resolution=60000,
    isotopes_file=None,
    logging_level="WARNING",
):
    """Calculate isotopologue correction factor for molecule.

    Takes pandas DataFrame and calculates isotopologue correction
    for molecule in molecules file, returns DataFrame with corrected values.
    Only C13 and N15 is supported as column labels right now e.g. 5C13

    Parameters
    ----------
    raw_data : pandas.DataFrame
        DataFrame of integrated lowest peaks per species vs time
    molecule_name : str
        Molecule name as in molecules_file.
    molecules_file : str or Path
        tab-separated file with name, formula and charge as rows
        e.g. Suc C4H4O3 -1
    molecule_formula : str
        Chemical formula as string.
        No spaces or underscores allowed.
        E.g. "C3H7O1"
    molecule_charge : int
        Charge as signed integer
    subset : list of str or False, optional
        List of column names to use for calculation
    exclude_col : list of str, optional
        Columns to ignore in calculation
    resolution_correction : bool, optional
        Run additonal correction for isotopologues overlaping
        due to low resolution. For example H02 and C13
    mz_calibration : float, optional
        mass-charge ratio of calibration point
    resolution : float, optional
        Resolution at calibration mz
    isotopes_file : Path or str, optional
        tab-separated file with element, mass, abundance and isotope as rows
        e.g. H 1.008 0.99 H01
    logging_level : str
        Logging is output to stderr
        Possible levels: "DEBUG", "INFO", "WARNING", "CRITICAL"
        Default level: "WARNING"

    Returns
    -------
    pandas.DataFrame
        Corrected data
    """
    logging.basicConfig(level=os.environ.get("LOGLEVEL", logging_level))
    dir_path = Path(__file__).parent
    if not isotopes_file:
        isotopes_file = dir_path / "isotopes.csv"
    if not molecules_file:
        molecules_file = dir_path / "metabolites.csv"
    if not subset:
        subset = raw_data.columns
        if exclude_col:
            subset = list(set(subset) - set(exclude_col))

    molecule_info = ip.MoleculeInfo.get_molecule_info(
        molecule_name, molecules_file, molecule_formula, molecule_charge, isotopes_file
    )
    subset = ip.LabelTuple(subset, molecule_info)
    res_corr_info = rc.ResolutionCorrectionInfo(
        resolution_correction, resolution, mz_calibration, molecule_info
    )
    if res_corr_info.do_correction:
        rc.warn_direct_overlap(subset, res_corr_info)
    clean_data = raw_data.fillna(0)
    data = correct_data(clean_data, subset, res_corr_info)
    return data


def correct_data(uncorrected_data, subset, res_corr_info):
    """Correct data based on labels in subset."""
    data = uncorrected_data.copy()
    for label1 in subset:
        corr = ip.calc_correction_factor(
            subset.molecule_info,
            label1,
        )
        assert corr >= 1, "Correction factor should be greater or equal 1"
        data[label1.as_string] = corr * data[label1.as_string]
        _logger.info(f"Correction factor {label1.as_string}: {corr}")
        for label2 in subset:
            if label1 >= label2:
                continue
            trans_prob = calc_transition_prob(label1, label2, res_corr_info)
            data[label2.as_string] = (
                data[label2.as_string] - trans_prob * data[label1.as_string]
            )
            # Ensure no negative values after correction
            data[label2.as_string] = data[label2.as_string].clip(lower=0)
            _logger.info(
                f"Transition prob {label1.as_string} -> {label2.as_string}: {trans_prob}"
            )
    return data


def calc_transition_prob(label1, label2, res_corr_info):
    """Calculate the probablity between two (un-)labelled isotopologues.

    Parameters
    ----------
    label1 : Label
        Type of isotopic label, e.g. Label("1N15")
    label2 : Label
        Type of isotopic label, e.g. Label("10C1301N15")
    res_corr_info : ResolutionCorrectionInfo
        Instance with resolution and mz calibration info.

    Returns
    -------
    float
        Transition probability
    """
    if label1 >= label2:
        return 0
    difference_labels = label2.subtract(label1)
    if res_corr_info.do_correction:
        trans_prob = rc.calc_indirect_overlap_prob(
            label1, label2, res_corr_info)
    else:  # Without resolution correction
        trans_prob = ip.calc_label_diff_prob(label1, difference_labels)
    return trans_prob



# -*- coding: utf-8 -*-
"""Isotopologue correction for MS data sets.

Modules:
    isotope_correction.py: main function
    isotope_probabilities.py: subroutines and helper functions
    resolution_correction.py: subroutines for resolution depend correction
"""
from picor.isotope_correction import calc_isotopologue_correction



"""Unit tests for resolution dependent isotope correction."""

import itertools
from numbers import Number
from pathlib import Path
import pytest
import unittest
import warnings

from picor.isotope_probabilities import MoleculeInfo, Label, LabelTuple
import picor.resolution_correction as rc


__author__ = "Jørn Dietze"
__copyright__ = "Jørn Dietze"
__license__ = "gpl3"


class TestResolutionCorrectionInfo(unittest.TestCase):
    """ResolutionCorrectionInfo class tests."""

    molecules_file = Path("tests/test_metabolites.csv")
    isotopes_file = Path("tests/test_isotopes.csv")
    molecule_info = MoleculeInfo.get_molecule_info(
        molecule_name="Test1",
        molecules_file=molecules_file,
        isotopes_file=isotopes_file,
    )

    def test_init_return_types(self):
        """Return correct types for attributes."""
        inst = rc.ResolutionCorrectionInfo(True, 100000, 600, self.molecule_info)
        self.assertIsInstance(inst.do_correction, bool)
        self.assertIsInstance(inst.resolution, Number)
        self.assertIsInstance(inst.mz_calibration, Number)
        self.assertIsInstance(inst.molecule_info, MoleculeInfo)
        self.assertIsInstance(inst.molecule_mass, Number)
        self.assertIsInstance(inst.min_mass_diff, Number)


class TestMassCalculations(unittest.TestCase):
    """Molecule mass and minimum mass difference."""

    molecules_file = Path("tests/test_metabolites.csv")
    isotopes_file = Path("tests/test_isotopes.csv")
    molecule_info = MoleculeInfo.get_molecule_info(
        molecule_name="Test1",
        molecules_file=molecules_file,
        isotopes_file=isotopes_file,
    )
    res_corr_info = rc.ResolutionCorrectionInfo(False, 60000, 200, molecule_info)

    def test_calc_min_mass_diff_result(self):
        """Minimal mass difference."""
        res = rc.ResolutionCorrectionInfo.calc_min_mass_diff(680, 2, 200, 50000)
        self.assertAlmostEqual(res, 0.029436, places=6)

    def test_calc_min_mass_diff_negative_mass(self):
        """Minimal mass difference."""
        with self.assertRaises(ValueError):
            rc.ResolutionCorrectionInfo.calc_min_mass_diff(-680, 2, 200, 50000)

    def test_is_overlap_true(self):
        """Overlapping isotopologues"""
        res = rc.is_isotologue_overlap(
            Label("5C13", self.molecule_info),
            Label("4C13 1H02", self.molecule_info),
            self.res_corr_info,
        )
        self.assertTrue(res)

    def test_is_overlap_false(self):
        """Non overlapping isotopologues"""
        res = rc.is_isotologue_overlap(
            Label("14C13", self.molecule_info),
            Label("14H02", self.molecule_info),
            self.res_corr_info,
        )
        self.assertFalse(res)

    def test_coarse_mass_difference_result(self):
        """Difference in nucleons."""
        res = rc.calc_coarse_mass_difference(
            Label("No label", self.molecule_info),
            Label("5C13 3N15 2H02", self.molecule_info),
        )
        self.assertEqual(res, 10)

    def test_coarse_mass_difference_bad_type(self):
        """Non Label as input."""
        label = Label("No label", self.molecule_info)
        non_label = {"C13": 12}
        label_list = [(label, non_label), (non_label, label), (non_label, non_label)]
        for la1, la2 in label_list:
            with self.subTest():
                with self.assertRaises(TypeError):
                    rc.calc_coarse_mass_difference(la1, la2)

    def test_fwhm_result(self):
        """Result with valid input."""
        mz_cal, mz, resolution = 200, 500, 50_000
        res = rc.ResolutionCorrectionInfo.fwhm(mz_cal, mz, resolution)
        self.assertAlmostEqual(res, 0.01581139)

    def test_fwhm_bad_input(self):
        """Result with valid input."""
        for mz_cal, mz, resolution in itertools.product([200, -200], repeat=3):
            if all(ele > 0 for ele in [mz_cal, mz, resolution]):
                # skip case in which all are positive
                continue
            with self.subTest():
                with self.assertRaises(ValueError):
                    rc.ResolutionCorrectionInfo.fwhm(mz_cal, mz, resolution)


class TestGenerateLabels(unittest.TestCase):
    """generate_labels."""

    molecules_file = Path("tests/test_metabolites.csv")
    isotopes_file = Path("tests/test_isotopes.csv")
    molecule_info = MoleculeInfo.get_molecule_info(
        molecule_name="Test1",
        molecules_file=molecules_file,
        isotopes_file=isotopes_file,
    )
    res_corr_info = rc.ResolutionCorrectionInfo(False, 60000, 200, molecule_info)

    def test_generate_labels_res(self):
        """Return correct result fo rsmall example."""
        res = list(rc.generate_labels(2, self.res_corr_info))
        res_corr = [
            Label(la, self.molecule_info)
            for la in [
                "1O18",
                "2O17",
                "1N151O17",
                "2N15",
                "1H021O17",
                "1H021N15",
                "2H02",
                "1C131O17",
                "1C131N15",
                "1C131H02",
                "2C13",
            ]
        ]
        self.assertListEqual(res, res_corr)

    def test_generate_labels_error(self):
        """Catch ValueError for generated labels too large for molecule."""
        mol = MoleculeInfo.get_molecule_info(
            molecule_formula="C2H8O1",
            molecule_charge=1,
            molecules_file=self.molecules_file,
            isotopes_file=self.isotopes_file,
        )
        res_corr_info = rc.ResolutionCorrectionInfo(False, 60000, 200, mol)
        res = len(list(rc.generate_labels(3, res_corr_info)))
        self.assertEqual(res, 9)


class TestOverlapWarnings(unittest.TestCase):
    """Overlap warnings."""

    molecules_file = Path("tests/test_metabolites.csv")
    isotopes_file = Path("tests/test_isotopes.csv")
    molecule_info = MoleculeInfo.get_molecule_info(
        molecule_name="Test1",
        molecules_file=molecules_file,
        isotopes_file=isotopes_file,
    )
    res_corr_info = rc.ResolutionCorrectionInfo(False, 60000, 200, molecule_info)

    def test_direct_overlap_warn(self):
        """Warning with overlapping labels."""
        with self.assertWarns(UserWarning):
            rc.warn_direct_overlap(
                LabelTuple(["4H02", "4C13"], self.molecule_info), self.res_corr_info
            )

    def test_direct_overlap_not_warn(self):
        """No warning with non-overlapping labels."""
        with warnings.catch_warnings():
            warnings.simplefilter("error")
            rc.warn_direct_overlap(
                LabelTuple(["3H02", "4C13"], self.molecule_info), self.res_corr_info
            )

    def test_indirect_overlap_warn(self):
        """Warning with indirectly overlapping labels."""
        res_corr_info_high = rc.ResolutionCorrectionInfo(
            False, 200000, 200, self.molecule_info
        )
        label_list = [
            LabelTuple(["3H02", "4C13"], self.molecule_info),
            LabelTuple(["1H02 1C13", "5C13"], self.molecule_info),
            LabelTuple(["2N15 1H02", "2N15 3C13"], self.molecule_info),
        ]
        for labels in label_list:
            with self.subTest():
                with self.assertWarns(UserWarning):
                    rc.warn_indirect_overlap(labels, res_corr_info_high)
